#acknowledgements img {
    height: 4em;
}

/* uni goettingen */
#acknowledgements img:first-of-type {
    margin: 1em 2em 1em 0;
}

/* dfg */
#acknowledgements img:nth-of-type(2) {
    margin: 1em 0 1em 0;
}

/* override no-wrap for inherited kwargs tables */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: normal;
}

p.rubric {
    border-bottom: 0em;
}

/* from https://stackoverflow.com/a/62451601 */
table.dataframe {
    display: block;
    max-width: -moz-fit-content;
    max-width: fit-content;
    overflow-x: auto;
}

table.dataframe th, table.dataframe td {
    border: 1px solid #c9c9c9;
    padding: 6px 13px;
}

table.dataframe tr:nth-child(even) {
    background-color: #fafafa;
}
